import express from 'express'
import fs from 'fs/promises'
import { exec as execAsync } from 'child_process'
import { objectMap, objectOmit } from '@dzeio/object-util'
import { getParams } from './gcodeUtils'
import util from 'util'
import { evaluate } from 'mathjs'

const exec = util.promisify(execAsync)
const server = express()

async function exists(file: string) {
	try {
		await fs.stat(file)
		return true
	} catch {
		return false
	}
}

// make the files dir

server.use((req, _, next) => {
    let data: Array<any> = []
    // req.setEncoding()
    req.on('data', function(chunk) {
       data.push(Buffer.from(chunk))
    });

    req.on('end', function() {
        req.body = Buffer.concat(data);
        next();
    });
});

const path = process.cwd()

server.get('/', (_, res) => {
	res.send('please read the README.md file...')
})

server.post('/', async (req, res) => {
	const file = (Math.random() * 1000000).toFixed(0)
	console.log('started processing new request', file)
	await fs.mkdir('files', { recursive: true })
	
	const overrides = objectOmit(req.query, 'algo', 'config')
	let config = `${path}/configs/` + (req.query?.config ?? 'config') + '.ini'
	if (!exists(config)) {
		console.log('request finished in error :(', file)
		res.status(403).send('Configuration file not existing')
		return
	}
	const stlPath = `${path}/files/${file}.stl`
	const gcodePath = `${path}/files/${file}.gcode`
	// console.log(stlPath, req.body)
	await fs.writeFile(stlPath, req.body, {
		encoding: null
	})
	// console.log(fs.statSync(stlPath).size, req.body.length)
	let additionnalParams = objectMap(overrides, (v, k) => `--${(k as string).replace(/_/g, '-')} ${v}`).join(' ')
	const slicer = process.env.SLICER_PATH ?? 'prusa-slicer'
	if (slicer.includes('prusa')) {
		additionnalParams += ' --export-gcode'
	}
	const cmd = `${slicer} ${stlPath} --load ${config} --output ${gcodePath} ${additionnalParams}`
	try {
		await exec(cmd, {
			timeout: 60000
		})
	} catch (e: any) {
		console.log('request finished in error :(', file)
		const line = e.toString()
		console.error(e)
		if (line.includes('Objects could not fit on the bed')) {
			await fs.rm(stlPath)
			res.status(403).send('Object is too large')
			return
		} else if (line.includes('No such file')) {
			await fs.rm(stlPath)
			res.status(403).send('Configuration file not existing')
			return
		} else if (line.includes('Unknown option')) {
			await fs.rm(stlPath)
			res.status(403).send('an override does not exists, please contact an administrator or refer to the documentation')
			return
		} else if (line.includes('ETIMEDOUT') || line.includes('Command failed')) {
			await fs.rm(stlPath)
			res.status(403).send('the file is taking too long to process :(')
			return
		}
		res.status(500).send(`Error while making the file:<br />${e.toString().replace(cmd, '<i>software</i>').replace(stlPath, `<i>file ${file}</i>`).replace(`${path}/configs/`, '').replace('.ini', '')}`)
		return
	}
	const gcode = await fs.readFile(gcodePath, 'utf-8')
	await fs.rm(stlPath)
	await fs.rm(gcodePath)
	const params = getParams(gcode)
	let price = -1
	if (req.query?.algo) {
		let algo = req.query.algo as string
		// objectLoop(params, (value, key) => {
		// 	if (typeof value !== 'number') {
		// 		return
		// 	}
		// 	while (algo.includes(key)) {
		// 		algo = algo.replace(key, value.toString())
		// 	}
		// })
		price = evaluate(algo, params)
	}
	res.json({
		price: parseFloat(price.toFixed(2)),
		...getParams(gcode),
		gcode
	})
	console.log('request successfull :)', file)
	// res.sendFile(gcodePath)
})

server.listen(3000, () => {
	console.log(`🚀 Server ready at localhost:3000`);
})
